/*
 *  eye.cc: prediction 
 *
 *  Created on: Nov. 1st, 2013
 *      Author: Tianbao Yang
 *
 * 
 *     birds fly on the cloud
 * 
 */
// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013 Tianbao Yang.
//  All Rights Reserved.
//*****************************************************************************
// Included Files
//*****************************************************************************
#include "cmd_line.h"
#include "inc_dual.h"
#include "data_set.h"
#include "objective.h"
#include <algorithm>
#include <functional>
#include <sstream>
#include <set>
#include <iomanip>
//#include <mpi.h>
//using namespace MPI;




int main(int argc, char** argv) {


  // -------------------------------------------------------------
  // ---------------------- Parse Command Line -------------------
  // -------------------------------------------------------------

  std::string test_filename;
  std::string model_filename;
  std::string pred_filename; 
  std::string test_label_filename;
  std::string class_filename; 

  std::string format;
  std::string dtype;

  uint dimension=0;
  uint batch=0;
  uint block=0;
  
  int normalize=0;
  int ls=2;
  int org=1; 
  int myclass;
  int wsize=0;
  int preds;

  double bias;

  // parse command line
  learning::cmd_line cmdline;
  cmdline.info("birds eye view for big data analytics (classification and regression)");
  cmdline.add_master_option("<test-data-file or test-data-folder>", &test_filename);
  cmdline.add("-norm", "normalize each data  or not (default = 0)", &normalize, 0);
  cmdline.add("-ls", "the loss function (1 (L1 hinge loss) or 2 (L2 hinge loss), 3 (logistic loss), 4(least square regression)) (default = 2)", &ls, 2);
  cmdline.add("-d", "input data dimension (only for binary data format) ", &dimension, 0);
  cmdline.add("-nb","number of bacthes of test data (default = number of procs)", &batch, 0);
  cmdline.add("-avgw", "evalue average model or not (default = no)", &wsize, 0);
  cmdline.add("-org","starting index of data (0, 1)  (default = 1)", &org, 1);
  cmdline.add("-fmt","data format (dense or sparse(libsvm)) (default = sparse)", &format, "sparse");
  cmdline.add("-bias", "adding bias or not ", &bias, 0.0);
  cmdline.add("-dtype","data type (txt or dat)  (default = txt)", &dtype, "txt");
  cmdline.add("-labelFile","name of test label foldername (for binary data option only,  assume files 1.txt, 2.txt) (default = noLabelFile)", &test_label_filename,"noLabelFile");
  //cmdline.add("-predFile","name of  predictions file/folder name (default = noPredFile)", &pred_filename,"noPredFile");
  cmdline.add("-preds","output predictions or not (default=0)", &preds,0);
  cmdline.add("-modelFile","name of model file (default = noModelFile)", &model_filename,"noModelFile");
  cmdline.add("-classFile","name of class file (for multiclass classification) (default = eagles_classes.txt)", &class_filename,"eagles_classes.txt");
  cmdline.add("-myclass","target class see output of birds.cc (default=1)", &myclass, 1);


  boost::mpi::environment env(argc, argv);
  boost::mpi::communicator COMM_WORLD;
  uint rank=COMM_WORLD.rank();
  uint size=COMM_WORLD.size();


  int rc = cmdline.parse(argc, argv);
  if (rc < 1) {
   if (rank==0)
   {
      cmdline.print_help();
      std::cout<<"number of argc="<<rc<<std::endl;
   }
   return EXIT_FAILURE;
  }
  if(model_filename=="noModelFile")
  {
  	  std::cerr<<"no model filename given "<<std::endl;
  	  return(EXIT_FAILURE); 
  }

  std::string loss_type;
  if(ls==1) loss_type="SVM(hinge)";
  else if(ls==2) loss_type="SVM(squared-hinge)";
  else if(ls==3) loss_type="LogReg";
  else if(ls==4) loss_type="LeastReg";

  // -------------------------------------------------------------
  // ---------------------- read the data ------------------------
  // ------------------------------------------------------------- 
    std::string task=(ls<=3)?"class":"regre";
    Dataset testData(format, task, dimension, normalize==1, bias); //(data_filename_r, format, task, dimension, normalize); 
    std::string data_filename_r = test_filename;
    std::string label_filename_r = test_label_filename;
      if(size>1){
        std::ostringstream convert;
        convert<<rank+org;
        data_filename_r = data_filename_r + "/" + convert.str();
        if(dtype!="") data_filename_r =data_filename_r + "."+dtype;
        if(test_label_filename!="noLabelFile") label_filename_r = label_filename_r + "/" + convert.str() +".txt";
      }    
    
    if(batch!=0 and block>batch) block=batch;
    if(block==0 or block>size) block=size;
    if(batch==0) {
       batch=block;
       if(rank==0) std::cerr<<"warning: check if #of baches is "<<batch<<std::endl;
       if(size==1) std::cerr<<"reading data from "<<data_filename_r<<std::endl;
    }
    if(block>batch) {
      if(rank==0) std::cerr<<"#of procs (np) > #of batches (nb); exit "<<std::endl; 
      return(EXIT_FAILURE);
    }
    if(rank==0) std::cout<<"reading data..."<<std::flush;
    if(block==batch){
           testData.Append(data_filename_r, label_filename_r);
      }else{
          uint num_per_block=floor((double)batch/(double)block);
          uint start_batch=rank*num_per_block, end_batch=(rank+1)*num_per_block;
          if(rank+1==size and end_batch<batch) end_batch=batch;
          for(uint b=start_batch; b<end_batch and b<batch; b++){
             std::ostringstream convert;
             convert<<b+org;
             data_filename_r = test_filename + "/" + convert.str();
             if(dtype!="") data_filename_r =data_filename_r + "."+dtype;
             if(test_label_filename!="noLabelFile")  label_filename_r=test_label_filename +"/"+convert.str()+".txt"; 
             testData.Append(data_filename_r, label_filename_r);             
           }
      }
      ulong ns=testData.size(); 
      uint ds=testData.dim();
      ulong n;
      all_reduce(COMM_WORLD, ns, n, std::plus<ulong>());
      all_reduce(COMM_WORLD, ds, dimension, boost::mpi::maximum<uint>());
 
    long testReadingTime=testData.readtime();
    if(rank==0){
      std::cout<< testReadingTime<<"s" <<std::endl<<std::flush;
    }  
 
  // reading model 
      std::ifstream model_file(model_filename.c_str());
      if (!model_file.good()) {
        std::cerr << "error w/ " << model_filename << std::endl;
        exit(EXIT_FAILURE);
      }
      uint num_models = 0;
      std::string buf;
      std::vector<WeightVector> Wc, AvgWc;  
      while (getline(model_file,buf)) {
       if (buf[0] == '#') break;
       std::istringstream is(buf);
       WeightVector W(is);
       if(W.dimension()!=dimension){
       	  std::cerr<<"model dimension!=data dimension"<<std::endl;
       	  return(EXIT_FAILURE);	     	   
       }
    	   
       Wc.push_back(W);
       num_models ++;
      }
           
      uint num_avg_models=0;
      if(wsize){
       while (getline(model_file,buf)){
        std::istringstream is(buf);
        WeightVector W(is);
        if(W.dimension()!=dimension){
        	  std::cerr<<"model dimension!=data dimension"<<std::endl;
        	  return(EXIT_FAILURE);	     	   
        }
        AvgWc.push_back(W);
        num_avg_models ++;
       }
      }
      model_file.close();
      if(num_models==0){
      	  std::cerr<<"no models "<<std::endl;
      	  return(EXIT_FAILURE);
       }
      if(wsize and num_models!=num_avg_models){
      	  std::cerr<<"#models not equal #avg models "<<std::endl;
      	  return(EXIT_FAILURE);	  
      }
      //reading class file  
      std::vector<int> Class;
      uint tc=num_models; 
        if(num_models>1){
           	std::ifstream cla(class_filename.c_str()); 
           	if (!cla.good()) {
           	    std::cerr << "error w/ " << class_filename << std::endl;
           	    std::cout << "using class 1..."<<num_models<<std::endl;
              	for(uint c=1; c<=num_models; c++)
              		Class.push_back(c);
           	    exit(EXIT_FAILURE);
           	}
               std::string line;
               while(std::getline(cla, line))
               {
                 std::istringstream in(line);
                 int label; 
                 in>>label;
                 Class.push_back(label);
               }
           	tc=Class.size();
           	cla.close();
           	if(tc!=num_models){
           	    std::cerr << "number of models != number of classes "<< std::endl;
           	    exit(EXIT_FAILURE);
           	}           		
        }
 
   long  iter_calc_obj_time, calc_obj_time;
   std::vector<double> test_loss(1+(wsize>0?1:0), 0), test_error(1+(wsize>0?1:0), 0), total_test_error(1+(wsize>0?1:0), 0);
   
   std::ofstream predf;
   std::string pred_filename_r = test_filename;
   if(size>1) {
       std::ostringstream convert;
       convert<<rank+org;
       pred_filename_r = pred_filename_r +  "/" + convert.str() +".preds.txt";
   } else 
     pred_filename_r = pred_filename_r + ".preds.txt";

   if(preds)	predf.open(pred_filename_r);
   if(num_models==1) objective(testData, iter_calc_obj_time,  test_loss, test_error, Wc[0], AvgWc[0], wsize, ls, myclass, predf.rdbuf());
   else 	         objective(testData, Class,  iter_calc_obj_time,  test_error, Wc, AvgWc, wsize, ls, predf.rdbuf());
   all_reduce(COMM_WORLD, test_error.data(), 2, total_test_error.data(), std::plus<double>());
   all_reduce(COMM_WORLD, iter_calc_obj_time, calc_obj_time, std::plus<long>());
   if(preds)	predf.close();

  
    if(rank==0){
            std::cout<<std::left<<"+--------------------------------------------+\n";
      std::cout<<"running summary: "<<loss_type<<"\n";
      std::cout<<std::setw(35)<<std::left<<"#examples(n):"<<n<<std::endl
               <<std::setw(35)<<std::left<<"#features(d):"<<dimension<<std::endl
               <<std::setw(35)<<std::left<<"#processors(-np):"<<size<<std::endl
               <<std::setw(35)<<std::left<<"evaluation-time(s):"<<double(calc_obj_time)/size<<std::endl
               <<std::setw(35)<<std::left<<"class label:"<<myclass<<std::endl
               <<std::setw(35)<<std::left<<"test error: last solution:"<<total_test_error[0]/n<<std::endl;
      if(wsize){
            std::cout<<std::setw(35)<<std::left<<"test error: avged solution:"<<total_test_error[1]/n<<std::endl;
       } 

      /*std::cout.width(10);
      std::cout.fill(' ');
      std::cout<<std::left;
      std::cout<<"+--------------------------------------------+\n";
      std::cout<<"running summary: "<<loss_type<<"\n";
      std::cout<<"examples(n)	dimensions(d)	time(s)\n"
               <<std::setw(16)<<n<<std::setw(16)<<dimension<<double(calc_obj_time)/size<<std::endl;
      if(wsize){
    	 std::cout<<"te_error(w)	te_eror(avgw)"<<std::endl
                 <<std::setw(16)<<total_test_error[0]/n<<total_test_error[1]/n<<std::endl;
      }
      else{ 
    	   std::cout<<"test_error"<<std::endl    
    	            <<std::setw(16)<<total_test_error[0]/n<<std::endl;
      }*/     

      std::cout<<"+--------------------------------------------+\n";
     }  
}















