// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013 Tianbao Yang.
//  All Rights Reserved.
//
// birds fly on the cloud
// main file implements te DisDCA for big data binary classification and regression
//*****************************************************************************
// Included Files
//*****************************************************************************
#include "cmd_line.h"
#include "inc_dual.h"
#include "data_set.h"
#include "objective.h"
#include <algorithm>
#include <functional>
#include <sstream>
#include <set>
#include <iomanip>
//#include <mpi.h>
//using namespace MPI;




int main(int argc, char** argv) {


  // -------------------------------------------------------------
  // ---------------------- Parse Command Line -------------------
  // -------------------------------------------------------------

  std::string data_filename;
  std::string test_filename;
  std::string model_filename;
  std::string obj_filename;
  std::string label_filename; 
  std::string test_label_filename;

  std::string sr;
  std::string ur;
  std::string format;
  std::string dtype;

  uint iter = 100;
  uint iiter=1;
  uint exam_per_iter = 1;
  uint iter_to_eval=0;
  uint dimension=0;
  uint batch=0;
  uint block=0;
  uint tebn=1;
  uint wsize=0;
  uint dual_out=0; 
  uint simple_out=0;
  
  int verb=0;
  int normalize=0;
  int ls=2;
  int org=1; 
  int monitor=1;
  
  //double eta_constant=1;
  double l1s=0;
  double tol=0;
  double lambda = 1.0;
  double bias=0;
  double margint=0; // tau  for l1 norm
  double margine=0; // eta/4  for l2 norm
  // parse command line
  learning::cmd_line cmdline;
  cmdline.info("birds for big data analytics (classification and regression)");
  cmdline.add_master_option("<data-file or data-folder>", &data_filename);
  cmdline.add("-lambda", "regularization parameter (default = 1/n)", &lambda, 0);
  cmdline.add("-margint", "the offset of margin ([0,1), default = 0)", &margint, 0);
  //cmdline.add("-margine", "the scale of margin and reg ([0,+infty), default = 0)", &margine, 0);
  cmdline.add("-T", "number of iterations for optimizing SVM (default = 1000)", &iter,1000);
  cmdline.add("-ws", "window size for averaging (default = no averaging)", &wsize, 0);
  cmdline.add("-m", "size of block for stochastic updates (default = 1000)", &exam_per_iter, 1000);
  cmdline.add("-verb", "display objective value or not (default = 0)", &verb, 0);
  cmdline.add("-tol", "tolerence for duality gap (default = None)", &tol, 0);
  cmdline.add("-norm", "normalize each data  or not (default = 0)", &normalize, 0);
  cmdline.add("-eval", "number of iterations to evalute objective (set to 0 or number of iterations to save time) (default = 0)", &iter_to_eval, 0);
  cmdline.add("-iiter", "number of iterations for optimizing sub-SVM (default = 1)", &iiter,1);
  //cmdline.add("-eta", "the scaling factor for the learning rate (default = 1)", &eta_constant, 1);
  //cmdline.add("-sr", "the rule for computing the scalar s (R1 or R2) (default = R1)", &sr, "R1");
  //cmdline.add("-ur", "the update rule (bat or inc) (default = inc)", &ur, "inc");
  cmdline.add("-ls", "the loss function (1 (L1 hinge loss) or 2 (L2 hinge loss), 3 (logistic loss), 4(least square regression)) (default = 2)", &ls, 2);
  cmdline.add("-l1", "the scalar before l1 norm (defualt =0, lambda/2*(|w|_2+2*l1*|w|_1))", &l1s, 0);
  cmdline.add("-d", "input data dimension (only for binary data format) ", &dimension, 0);
  cmdline.add("-bias", "adding bias or not ", &bias, 0.0);
  cmdline.add("-nb","number of bacthes of data (default = number of procs)", &batch, 0);
  //cmdline.add("-nb","number of blocks of data (default = number of procs)", &block, 0);
  cmdline.add("-nt","number of bacthes of test data (default = 1)", &tebn, 1);
  cmdline.add("-org","starting index of data (0, 1)  (default = 1)", &org, 1);
  cmdline.add("-track", "track test error or not (default = 0)", &monitor, 0);
  cmdline.add("-fmt","data format (dense or sparse(libsvm)) (default = sparse)", &format, "sparse");
  cmdline.add("-dtype","data type (txt or dat)  (default = txt)", &dtype, "txt");
  cmdline.add("-labelFile","name of label file/folder name (for binary data option only,  assume files 1.txt, 2.txt) (default = noLabelFile)", &label_filename,"noLabelFile");
  cmdline.add("-testFile","name of test data file/folder name (default = noTestFile)", &test_filename,"noTestFile");
  cmdline.add("-testlabelFile","name of test label file/folder name (for binary data option only, assume files are 1.txt, 2.txt) (default = noLabelFile)", &test_label_filename,"noLabelFile");
  cmdline.add("-modelFile","name of model file (default = noModelFile)", &model_filename,"noModelFile");
  cmdline.add("-objFile","name of objective file (default = noObjFile)", &obj_filename, "noObjFile");
  cmdline.add("-dualout","output dual variables or not (default = 0)", &dual_out, 0);
  cmdline.add("-simpleout","if true just out training and testing error (default = 0)", &simple_out, 0);
  

  boost::mpi::environment env(argc, argv);
  boost::mpi::communicator COMM_WORLD;
  uint rank=COMM_WORLD.rank();
  uint size=COMM_WORLD.size();


  int rc = cmdline.parse(argc, argv);
  if (rc < 1) {
   if (rank==0)
   {
      cmdline.print_help();
      std::cout<<"number of argc="<<rc<<std::endl;
   }
   return EXIT_FAILURE;
  }


  /*if(ur!="inc" and ur!="bat"){
  	  std::cerr<<"ur = inc or bat"<<std::endl;
  	    return(EXIT_FAILURE);
    }
  if(sr!="R1" and ur!="R2"){
  	  std::cerr<<"sr = R1 or R2"<<std::endl;
  	    return(EXIT_FAILURE);
    }*/
  if(ls<=2 and margint>=1){
   if (rank==0) 
    {
      std::cerr<<"margin offset larger than 1 not allowed for classification"<<std::endl;
      cmdline.print_help();
    }
   return EXIT_FAILURE;
  }
 
  if(monitor!=0 and test_filename=="noTestFile")
  {
	  std::cerr<<"no test file for tracking"<<std::endl; 
	  monitor=0;
  }
  std::string loss_type;
  if(ls==1) loss_type="SVM(hinge)";
  else if(ls==2) loss_type="SVM(squared-hinge)";
  else if(ls==3) loss_type="LogReg";
  else if(ls==4) loss_type="LeastReg";

  // -------------------------------------------------------------
  // ---------------------- read the data ------------------------
  // -------------------------------------------------------------
  std::string data_filename_r = data_filename;
  std::string label_filename_r = label_filename;
   
  if(size>1){
    std::ostringstream convert;
    convert<<rank+org;
    data_filename_r = data_filename + "/" + convert.str();
    if(dtype!="") data_filename_r =data_filename_r + "."+dtype;
    if(label_filename!="noLabelFile") label_filename_r = label_filename + "/" + convert.str() +".txt";
  }
  
  std::string task=(ls<=3)?"class":"regre";
     
  trainDataset Data(format, task, dimension, normalize==1, bias); //(data_filename_r, format, task, dimension, normalize); 
  
  
  if(batch!=0 and block>batch) block=batch;
  if(block==0 or block>size) block=size;
  if(batch==0){
     batch=block;
     if(rank==0) std::cerr<<"warning: check if  #of bacthes (nb) of data is "<<block<<std::endl;
  }
  if(block>batch) {
    if(rank==0) std::cerr<<"#of procs (np) > #of batches (nb); exit "<<std::endl; 
    return(EXIT_FAILURE);
  }
  if(rank==0 and simple_out==0) std::cout<<"reading data..."<<std::flush;
  if(block==batch){
       Data.Append(data_filename_r, label_filename_r);
  }else{
      uint num_per_block=floor((double)batch/(double)block);
      uint start_batch=rank*num_per_block, end_batch=(rank+1)*num_per_block;
      if(rank+1==size and end_batch<batch) end_batch=batch;
      for(uint b=start_batch; b<end_batch and b<batch; b++){
         std::ostringstream convert;
         convert<<b+org;
         data_filename_r = data_filename + "/" + convert.str();
         if(dtype!="") data_filename_r =data_filename_r + "."+dtype;
         if(label_filename!="noLabelFile")  label_filename_r=label_filename +"/"+convert.str()+".txt"; 
         Data.Append(data_filename_r, label_filename_r);             
       }
  }
  
  
  ulong ns=Data.size();
  uint ds=Data.dim();
  double Rs=Data.mR();
  double Sqys=Data.mS();
  long readingTime=Data.readtime();
  
  ulong n;
  double R; 
  double Sqy; 
  all_reduce(COMM_WORLD, ns, n, std::plus<ulong>());
  all_reduce(COMM_WORLD, Rs, R, boost::mpi::maximum<double>());
  all_reduce(COMM_WORLD, ds, dimension, boost::mpi::maximum<uint>());
  if(ls==4){
	 all_reduce(COMM_WORLD, Sqys, Sqy, std::plus<double>());
     Sqy /=n;
  }

  if(rank==0 and (R==0 || ns==0 || n==0)){
    std::cerr<<"something wrong in the data"<<std::endl;
    std::cerr<<"max square norm="<<R<<",ns="<<ns<<",n="<<n<<std::endl;
    if(size==1){
    	  std::cerr<<"hint: ensure your data-file is a filename or use -np #of batches"<<std::endl;
      }
    return(EXIT_FAILURE);
  }  
  
    Dataset testData(format, task, dimension - (bias!=0), normalize==1, bias); //(data_filename_r, format, task, dimension, normalize); 
    long testReadingTime=0;
    if (test_filename != "noTestFile" && rank==0) {
  	  if(tebn==1){
         testData.Append(test_filename, test_label_filename);
  	  }else{
  		 for(uint b=0; b<tebn; b++){
  		      std::string test_filename_r, test_label_filename_r;
  		      std::ostringstream convert;
  		      convert<<b+org;
  		      test_filename_r = test_filename + "/" + convert.str();
  		      if(dtype!="") test_filename_r=test_filename_r + "."+dtype;
  		      if(test_label_filename!="noLabelFile")  test_label_filename_r=test_label_filename +"/"+convert.str()+".txt"; 
  		      testData.Append(test_filename_r, test_label_filename_r);
  		 }
  	  }
  	testReadingTime=testData.readtime();  
   }else{
      testReadingTime = 0;
    }
    long my_readingTime = readingTime + testReadingTime;
    
    if(rank==0 and simple_out==0){
      std::cout<< my_readingTime<< "s" <<std::endl<<std::flush;
    }  
 
  std::set<int> uLabels;
  int myclass=1;
  if(task=="class")  uLabels=Data.mclass();
  if(uLabels.size()>2) {
    if(rank==0) std::cout<<"#of classes > 2, birds is small,  use eagles please\n"; 
    return(EXIT_SUCCESS);}
  else{
	 myclass=*std::max_element(uLabels.begin(), uLabels.end()); 
  }

  
  // choose a random seed
   srand(1);
   // ---------------start loop here--------------------------------
   // --------------------------------------------------------------
   // -------------------------------------------------------------
   // -------------------------------------------------------------
   long train_calc_obj_time[2]={0,0}, iter_calc_obj_time, total_train_calc_obj_time[2]={0,0};

   //to output the objective using both w and averaged w
   std::vector<double> phi_loss_error_norm(7, 0.0);
   std::vector<double> total_phi_loss_error(5, 0.0);
   std::vector<std::vector<double> > obj_vec;

   std::vector<double> test_loss(1+(wsize>0?1:0), 0), test_error(1+(wsize>0?1:0), 0);

   WeightVector W(dimension), Wr(dimension),  AvgW(dimension); // define global primal variable, local primal variable, and global averaged variable
   std::vector<double> Alpha(ns, 0.0); // dual variable 
  

   if(lambda==0) lambda=1.0/n;
   double old_lambda=lambda, old_margint = margint;
   if(ls==1 and margine>0) 
   { 
     ls=2; lambda *=margine; margint = (margine + margint -1)/margine;
   } else if(ls==2) {
     lambda *=margine+1;  margint = (margine + margint)/(margine+1); 
   }
   if(exam_per_iter==0 or exam_per_iter>ns) exam_per_iter=ns;
   if((tol>0 and iter_to_eval==0) or (monitor and iter_to_eval==0)) iter_to_eval = iter; // every iter  to evaluate the objective

   double coeff=1.0*size/(n*lambda);
   //double beta = exam_per_iter;
   double conj_val=0;
   double st_obj=0;
   
   if(ls<=2){// L1-SVM, L2-SVM; consider the margin offset
	 std::vector<double> obj(2+(wsize>0?1:0), 1.0 - margint); 
     obj_vec.push_back(obj);
     st_obj=1.0 - margint; 
   }
   else if(ls==3){//logistic regression
    std::vector<double> obj(2+(wsize>0?1:0), log(2)); 
 	obj_vec.push_back(obj);
 	st_obj=log(2);
   }
   else if(ls==4){//least square regression
	 std::vector<double> obj(2+(wsize>0?1:0), Sqy); 
     obj_vec.push_back(obj);
     st_obj=Sqy;
   }
   
   long startTime = get_runtime(), endTime;

   
   std::vector<ulong> index;
   for(ulong i=0; i<ns; i++) index.push_back(i); // handle the difference size of different portions of data
   ulong shift=0; // use to index the element for updating
   uint count=0; // count for counting the number of averaged W
   double dual_gap=0;
   double *wr, *w;

   //---------------Main Loop---------------------------------
   uint i=0, iter_incre=iter;
   if(verb and rank==0) std::cout<<"optimize model..."<<std::endl<<std::flush; 
   while(i<iter){

     //if(ur=="inc") 
     dual_update(ls, Data, W, Wr, Alpha, coeff,  phi_loss_error_norm, iiter, exam_per_iter, index, shift, myclass, margint);	
     wr= Wr.make_my_a_one();
     w = W.make_my_a_one();
     all_reduce(COMM_WORLD, wr, dimension, w, std::plus<double>());    
 
  	 //all_reduce(COMM_WORLD, Wr, W, std::plus<WeightVector>()); 
 	 W.scale(1.0/size);

 	 if(l1s>0.0) conj_val = W.proximal_l1(l1s);
 	 if(wsize and iter-i<=wsize){
 	  AvgW.scale(count); AvgW.add(W,1.0); AvgW.scale(1.0/(count+1));
 	  count++;
     }
 	 
	// compute the objective for monitoring 
 	if(iter_to_eval && (i+1) % iter_to_eval==0 ){
 		long iter_calc_start_time=get_runtime();
 	        objective(Data, iter_calc_obj_time, phi_loss_error_norm, W, AvgW, 1, count, ls, l1s, myclass, margint);
 		all_reduce(COMM_WORLD, phi_loss_error_norm.data(), 5, total_phi_loss_error.data(), std::plus<double>());
 		total_phi_loss_error[0] /=n; //average of dual phi
 		
 		std::vector<double> obj; 
	    total_phi_loss_error[1] /=n; //average of primal loss
	    total_phi_loss_error[3] /=n; //average of error 
	    obj.push_back(phi_loss_error_norm[5] * lambda / 2.0 + total_phi_loss_error[1]);
	    
	    if(wsize){ 
	    	if(count==0) obj.push_back(st_obj);
	        else { total_phi_loss_error[2] /=n; total_phi_loss_error[4] /=n; 
	               obj.push_back(phi_loss_error_norm[6] * lambda / 2.0 + total_phi_loss_error[2]);
	             }
	    }

 		if(l1s==0.0) dual_gap=obj[0] - total_phi_loss_error[0] + phi_loss_error_norm[5] * lambda/2.0;
 		else         dual_gap=obj[0] - total_phi_loss_error[0] +  conj_val * lambda;
 		if(dual_gap<0) dual_gap=1.0e-15;
 	    obj.push_back(dual_gap);
 		obj_vec.push_back(obj);
 		
 		if(monitor){
 			objective(testData, iter_calc_obj_time,  test_loss, test_error, W, AvgW, 0, ls, myclass);
 	    	if(monitor and rank==0)  std::cout<<"iter = "<<i+1<<"  test error = "<<test_error[0]/testData.size()<<std::endl;

 		}

 		
 		long iter_calc_end_time=get_runtime();
 		iter_calc_obj_time = iter_calc_end_time - iter_calc_start_time;
 	   }else iter_calc_obj_time=0;

 	   endTime=get_runtime();
 	   train_calc_obj_time[0] +=endTime - startTime - iter_calc_obj_time;
 	   train_calc_obj_time[1] +=iter_calc_obj_time;
 	   startTime = get_runtime(); 	 
    
 	   i++;
       if(i==iter){
 	   if(tol>0 and dual_gap>tol){
 	      iter +=iter_incre;
          if(verb and rank==0 and i%iter_to_eval==0){
        	  std::cout<<"iter = "<<i<<" dual gap = "<<dual_gap;
        	  std::cout<<std::endl;
          }
 	   }else if(dual_gap<tol)
        {iter=i; break;}
 	    else if(tol==0) break;
       } 
 }// end for(int i=0; ...)
 if(verb and rank==0) std::cout<<"optimize model done!"<<std::endl<<std::flush; 
 
  
    endTime=get_runtime();
    train_calc_obj_time[0] += endTime - startTime;
    all_reduce(COMM_WORLD, train_calc_obj_time, 2, total_train_calc_obj_time, std::plus<long>());
    if(iter_to_eval ==0 ){
     	    objective(Data, iter_calc_obj_time, phi_loss_error_norm, W, AvgW, 1, count, ls, l1s, myclass, margint);
     		all_reduce(COMM_WORLD, phi_loss_error_norm.data(), 5, total_phi_loss_error.data(), std::plus<double>());
     		total_phi_loss_error[0] /=n; //average of dual phi
     		
     		std::vector<double> obj; 
    	    total_phi_loss_error[1] /=n; //average of primal loss
    	    total_phi_loss_error[3] /=n; //average of error 
    	    obj.push_back(phi_loss_error_norm[5] * lambda / 2.0 + total_phi_loss_error[1]);
    	    
    	    if(wsize){ 
    	    	if(count==0) obj.push_back(st_obj);
    	        else { total_phi_loss_error[2] /=n; total_phi_loss_error[4] /=n; 
    	               obj.push_back(phi_loss_error_norm[6] * lambda / 2.0 + total_phi_loss_error[2]);
    	             }
    	    }

     		if(l1s==0.0) dual_gap=obj[0] - total_phi_loss_error[0] + phi_loss_error_norm[5] * lambda/2.0;
     		else         dual_gap=obj[0] - total_phi_loss_error[0] +  conj_val * lambda;
     		if(dual_gap<0) dual_gap=1.0e-12;
     	    obj.push_back(dual_gap);
     		obj_vec.push_back(obj);
     	   } 
 

    lambda =old_lambda; margint = old_margint;
 
    if(obj_filename != "noObjFile" && iter_to_eval && rank==0){
      if(verb and rank==0) std::cout<<"print obj..."<<std::endl<<std::flush; 
      std::ofstream obj(obj_filename.c_str());
      for(uint it=0; it<obj_vec.size(); it++){
        copy(obj_vec[it].begin(), obj_vec[it].end(), std::ostream_iterator<double>(obj, " "));
        obj<<std::endl;
      }
      obj.close();
      if(verb and rank==0) std::cout<<"print obj done!"<<std::endl<<std::flush; 
     }
    
     if (test_filename != "noTestFile" and rank==0){
         if(verb and rank==0) std::cout<<"evalute test..."<<std::endl<<std::flush; 
    	 objective(testData, iter_calc_obj_time,  test_loss, test_error, W, AvgW, wsize, ls, myclass);
         if(verb and rank==0) std::cout<<"evalute test done!"<<std::endl<<std::flush; 
     }
     
     if (model_filename != "noModelFile" and rank==0) {
       if(verb and rank==0) std::cout<<"print model..."<<std::endl<<std::flush; 
       std::ofstream model_file(model_filename.c_str());
       if (!model_file.good()) {
         std::cerr << "error w/ " << model_filename << std::endl;
         exit(EXIT_FAILURE);
       }
       W.print(model_file);       
       if(wsize) { model_file<<"#---------avg model--------"<<std::endl; AvgW.print(model_file);}
       model_file.close();
       if(verb and rank==0) std::cout<<"print model done!"<<std::endl<<std::flush; 

     }
 
    if (dual_out) {
     if (verb and rank==0) std::cout<<"print dual ..."<<std::endl<<std::flush; 
     std::string dual_filename_r = data_filename;
     if(size>1){
       std::ostringstream convert_r, convert_m, convert_l, convert_e, convert_b;
       convert_r<<rank+org;
       convert_m<<margint; 
       convert_l<<lambda;
       convert_e<<margine;
       convert_b<<bias;
       dual_filename_r = dual_filename_r + "/" + convert_r.str() + ".lambda." + convert_l.str() + ".margint." + convert_m.str()+ ".margine." + convert_e.str() + ".dual.txt";
     } else
       dual_filename_r = dual_filename_r + ".dual.txt";  
     std::ofstream dual(dual_filename_r);
     copy(Alpha.begin(), Alpha.end(), std::ostream_iterator<double>(dual, "\n"));
    }
  
    if(rank==0){
     if(simple_out==1){
       std::cout<<std::setw(35)<<std::left<<total_phi_loss_error[3];
       if(test_filename!="noTestFile"){
         std::cout<<std::setw(35)<<std::left<<test_error[0]/testData.size();
        }
       std::cout<<std::endl;
     } else {
      std::cout<<std::left<<"+--------------------------------------------+\n";
      std::cout<<"running summary: "<<loss_type<<"\n";
      std::cout<<std::setw(35)<<std::left<<"#examples(n):"<<n<<std::endl
               <<std::setw(35)<<std::left<<"#features(d):"<<dimension<<std::endl
               <<std::setw(35)<<std::left<<"#examples-per-iteration(-m):"<<exam_per_iter<<std::endl
               <<std::setw(35)<<std::left<<"#processors(-np):"<<size<<std::endl
               <<std::setw(35)<<std::left<<"#iterations(-T):"<<iter<<std::endl
               <<std::setw(35)<<std::left<<"training-time(s):"<<total_train_calc_obj_time[0]/size<<std::endl
               <<std::setw(35)<<std::left<<"evaluation-time(s):"<<total_train_calc_obj_time[1]/size<<std::endl
               <<std::setw(35)<<std::left<<"regularization param(-lambda):"<<lambda<<std::endl
               <<std::setw(35)<<std::left<<"margin offset param(-margint):"<<margint<<std::endl
               <<std::setw(35)<<std::left<<"norm: last solution:"<<phi_loss_error_norm[5]<<std::endl
               <<std::setw(35)<<std::left<<"loss: last solution:"<<total_phi_loss_error[1]<<std::endl
               <<std::setw(35)<<std::left<<"primal objective:"<<obj_vec.back()[0]<<std::endl;
      if(l1s>0)
        std::cout<<std::setw(35)<<std::left<<"dual objective:"<<total_phi_loss_error[0] -conj_val*lambda<<std::endl;
      else
        std::cout<<std::setw(35)<<std::left<<"dual objective:"<<total_phi_loss_error[0] - phi_loss_error_norm[5]*lambda/2.0<<std::endl;
      std::cout<<std::setw(35)<<std::left<<"duality gap:"<<dual_gap<<std::endl
               <<std::setw(35)<<std::left<<"class label:"<<myclass<<std::endl
               <<std::setw(35)<<std::left<<"train error: last solution:"<<total_phi_loss_error[3]<<std::endl;
      if(test_filename!="noTestFile"){
         std::cout<<std::setw(35)<<std::left<<"test error: last solution:"<<test_error[0]/testData.size()<<std::endl;
      }
      if(wsize){
        std::cout<<std::setw(35)<<std::left<<"norm: avged solution:"<<phi_loss_error_norm[6]<<std::endl
                 <<std::setw(35)<<std::left<<"loss: avged solution:"<<total_phi_loss_error[2]<<std::endl
                 <<std::setw(35)<<std::left<<"train error: avged solution:"<<total_phi_loss_error[4]<<std::endl;
        if(test_filename!="noTestFile")
            std::cout<<std::setw(35)<<std::left<<"test error: avged solution:"<<test_error[1]/testData.size()<<std::endl;
       }
      /*std::cout<<"examples(n)	dimensions(d)	lambda\n"
               <<std::setw(16)<<n<<std::setw(16)<<dimension<<lambda<<std::endl;
               //<<"--------------------------------------------\n";
      std::cout<<"process(K)	exam_per(m)	iter_to_eval\n"
               <<std::setw(16)<<size<<std::setw(16)<<exam_per_iter<<iter_to_eval<<std::endl;
               //<<"--------------------------------------------\n";          		   
      std::cout <<"Iters		Train-time	Eval-Time\n"
                <<std::setw(16)<<iter<<std::setw(16)<<total_train_calc_obj_time[0]/size
                <<total_train_calc_obj_time[1]/size<<std::endl;
                //<<"--------------------------------------------\n";
      std::cout<<"Norm		Loss		Obj\n"
                <<std::setw(16)<<phi_loss_error_norm[5]<<std::setw(16)<<total_phi_loss_error[1]
                <<obj_vec.back()[0]<<std::endl;
                //<<"--------------------------------------------\n";
      std::cout<<"primal		dual		gap\n";
      if(l1s>0) std::cout <<std::setw(16)<<obj_vec.back()[0]<<std::setw(16)<< total_phi_loss_error[0] -conj_val*lambda
                        <<dual_gap<<std::endl;
      else    std::cout <<std::setw(16)<<obj_vec.back()[0]<<std::setw(16)<<total_phi_loss_error[0] - phi_loss_error_norm[5]*lambda/2.0
                        <<dual_gap<<std::endl;
      //std::cout<<"--------------------------------------------\n";
      std::cout<<"class		gap		train_error\n"
                <<std::setw(16)<<myclass<<std::setw(16)<<dual_gap<<total_phi_loss_error[3]<<std::endl;
                //<<"--------------------------------------------\n";
      if(wsize){
      std::cout<<"\nNorm(avgw)	Loss		train_error\n"
               <<std::setw(16)<< phi_loss_error_norm[6]<<std::setw(16)<< total_phi_loss_error[2]
               <<total_phi_loss_error[4]<<std::endl;
               //<<"--------------------------------------------\n";
      }
      if(test_filename!="noTestFile"){
         if(wsize) std::cout<<"te_loss(w)	te_error(w)	te_eror(avgw)\n";
         else      std::cout<<"test_loss 	test_error\n";         
    	 if(wsize) std::cout<<std::setw(16)<<test_loss[0]/testData.size()<<std::setw(16)<<test_error[0]/testData.size()
    	                    <<test_error[1]/testData.size()<<std::endl;
    	 else std::cout<<std::setw(16)<<test_loss[0]/testData.size()<<std::setw(16)<<test_error[0]/testData.size()<<std::endl;
      } */    
      std::cout<<"+--------------------------------------------+\n"<<  std::endl;
     }
    }  
}















