// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013 Tianbao Yang.
//  All Rights Reserved.
//============================================================================================
// File Name: data_set.cc
// implemetns the member functions for  Dataset class
//==============================================================================================
#include "data_set.h"


  void Dataset::ReadData(std::string& data_filename){
	  
	  // Start a timer
	  long startTime = get_runtime();

	  // OPEN DATA FILE
	  // =========================
	  std::ifstream data_file(data_filename.c_str());
	  if (!data_file.good()) {
	    std::cerr << "error w/ " << data_filename << std::endl;
	    exit(EXIT_FAILURE);
	  }


	  // Read SVM-Light or Dense data file
	  // ========================
	  std::string buf;
	  while (getline(data_file,buf)) {
	    // ignore lines which begin with #
	    if (buf[0] == '#') continue;
	    // Erase what comes after #
	    size_t pos = buf.find('#');
	    if (pos < buf.size()) {
	      buf.erase(pos);
	    }
	    // replace ':' with white space
	    int n=0;
	    for (size_t pos=0; pos < buf.size(); ++pos)
	     if(format=="sparse"){
	       if (buf[pos] == ':') {
		      n++; buf[pos] = ' ';
	       }
	     }else{
	       if (buf[pos] == ' ') n++;
	     }
	    	
	    // read from the string
	    std::istringstream is(buf);
	    
	    if(task=="class"){
	      int label = 0;
	      is >> label;
	      Labels.push_back(label);
	    }else{
	      double label=0.0;
	      is >> label;
	      Targets.push_back(label);
	    }
	    
	    if(format=="dense"){
	      dense_vector instance(is,n);
	      //double isn=instance.snorm();
	      if(normalize){
	       //instance.scale(1.0/std::sqrt(isn));
	       instance.normalize();
	      }
	      densedata.push_back(instance);
	      if(dimension==0) dimension=n;
	      else if((uint)n!=dimension)
	  	  {
	  	    std::cerr << "dimension not consistent at line "<<num_examples+1<< " in file "<<data_filename<< std::endl;
	        exit(EXIT_FAILURE);
	  	  }  	  
	    }else{
	      simple_sparse_vector instance(is,n);
	      double isn=instance.snorm();
	      if(normalize) instance.scale(1.0/std::sqrt(isn));
	      sparsedata.push_back(instance);
		  uint cur_max_ind = instance.max_index() + 1;
		  if (cur_max_ind > dimension) dimension = cur_max_ind;
	    }
	    num_examples++;
	  }

	  data_file.close();


	#ifdef nodef
	  std::cerr << "num_examples = " << num_examples
		    << " dimension = " << dimension << std::endl;
	#endif


	  // update timeline
	  readingTime += get_runtime() - startTime;
  }
  
  
  
  void Dataset::ReadBinaryData(std::string& data_filename){
	  long startTime = get_runtime();
	  std::ifstream data_file(data_filename.c_str(), std::ios::in | std::ios::binary);
	  if (!data_file.good()) {
		    std::cerr << "error w/ " << data_filename << std::endl;
		    exit(EXIT_FAILURE);
	  } 
	   data_file.seekg (0, data_file.end);
	   long total_size = data_file.tellg();
	   data_file.seekg (0, data_file.beg);

	   int unit_size =sizeof(float);	   
	   for(int j=0; j<total_size/(dimension*unit_size); j++){
	  	 std::ostringstream buf; 	  	
	  	 std::vector<float> vec(dimension,0.0);
	  	 data_file.read((char*)vec.data(), unit_size*dimension);
	  	 if (data_file.good()){
	  	      if(format=="sparse")
	  	      	std::copy(vec.begin(), vec.end(), std::ostream_iterator<float>(buf, " "));
	  	      else{
	  	    	dense_vector instance; 
	  	        instance.initialize(dimension, vec);
	  	        //double isn=instance.snorm();
	  	     	if(normalize) {
	  	          instance.normalize(); //instance.scale(1.0/sqrt(isn));
	  	     	}
	  	     	densedata.push_back(instance);
	  	      }
	  	 }
	  	 else { std::cout << "read data "<<j+1<<" failed"<<std::endl; exit(EXIT_FAILURE);}            

	     //std::cout<<buf.str()<<std::endl;
	  	 if(format=="sparse"){
	  	   std::istringstream is(buf.str());
	  	   simple_sparse_vector instance(is, dimension, false);
    	   double isn=instance.snorm();
	       if(normalize) instance.scale(1.0/sqrt(isn));
	  	   sparsedata.push_back(instance);
	  	 }
	     num_examples ++;
	   }
	   
	   data_file.close();
	   
       #ifdef nodef
          std::cerr << "num_examples = " << num_examples
	                << " dimension = " << dimension << std::endl;
       #endif
	   readingTime += get_runtime() - startTime;

  }
  
  
  void Dataset::ReadLabels(std::string& label_filename){
	     long startTime = get_runtime();
		 std::ifstream label_file(label_filename.c_str());
		  if (!label_file.good()) {
		    std::cerr << "error w/ " << label_filename << std::endl;
		    exit(EXIT_FAILURE);
		  }
	    std::string line;
	    ulong n=0;
	    while(std::getline(label_file, line))
	    {
	      std::istringstream in(line);
	      if(task=="class"){
	        int label; 
	        in>>label;
	        Labels.push_back(label);
	      }else{
	    	double label;
	    	in >> label;
	    	Targets.push_back(label);
	      }
	    }
	    if(task=="class") n=Labels.size();
	    else  n = Targets.size();
	    if(n!=num_examples){
		    std::cerr << "error data size "<<num_examples<<" != label size "<<n<<" in "<<label_filename<< std::endl;
		    exit(EXIT_FAILURE);
	    }
	    	
	    label_file.close();
	   	// update timeline
	   	readingTime += get_runtime() - startTime;
  }
  
  
  void trainDataset::ReadData(std::string& data_filename){
	  
	  // Start a timer
	  long startTime = get_runtime();

	  // OPEN DATA FILE
	  // =========================
	  std::ifstream data_file(data_filename.c_str());
	  if (!data_file.good()) {
	    std::cerr << "error w/ " << data_filename << std::endl;
	    exit(EXIT_FAILURE);
	  }


	  // Read SVM-Light or Dense data file
	  // ========================
	  std::string buf;
	  while (getline(data_file,buf)) {
	    // ignore lines which begin with #
	    if (buf[0] == '#') continue;
	    // Erase what comes after #
	    size_t pos = buf.find('#');
	    if (pos < buf.size()) {
	      buf.erase(pos);
	    }
	    // replace ':' with white space
	    int n=0;
	    for (size_t pos=0; pos < buf.size(); ++pos)
	     if(format=="sparse"){
	       if (buf[pos] == ':') {
		      n++; buf[pos] = ' ';
	       }
	     }else{
	       if (buf[pos] == ' ') n++;
	     }
	    	
	    // read from the string
	    std::istringstream is(buf);
	    
	    if(task=="class"){
	      int label = 0;
	      is >> label;
	      S +=label*label;
	      Labels.push_back(label);
	      uLabels.insert(label);
	    }else{
	      double label=0.0;
	      is >> label;
	      S +=label*label;
	      Targets.push_back(label);
	    }
	    
	    if(format=="dense"){
	      dense_vector instance(is,n);
	      double isn=instance.snorm();
	      if(normalize){
	        //instance.scale(1.0/sqrt(isn));
	        instance.normalize();
	        R=1.0;
	        Q.push_back(1.0);
	        }else{
	         if(isn>R) R=isn;
	         Q.push_back(isn);
	      }
	      densedata.push_back(instance);
	      if(dimension==0) dimension=n;
	      else if((uint)n!=dimension)
	  	  {
	  	    std::cerr << "dimension not consistent at line "<<num_examples+1<< " in file "<<data_filename<< std::endl;
	        exit(EXIT_FAILURE);
	  	  }  	  
	    }else{
	      simple_sparse_vector instance(is,n);
	      double isn=instance.snorm();
	      if(normalize){
	        instance.scale(1.0/sqrt(isn));
	        R=1.0;
	        Q.push_back(1.0);
	        }else{
	         if(isn>R) R=isn;
	         Q.push_back(isn);
	      }
	      sparsedata.push_back(instance);
		  uint cur_max_ind = instance.max_index() + 1;
		  if (cur_max_ind > dimension) dimension = cur_max_ind;
	    }
	    num_examples++;
	  }

	  data_file.close();


	#ifdef nodef
	  std::cerr << "num_examples = " << num_examples
		    << " dimension = " << dimension << std::endl;
	#endif


	  // update timeline
	  readingTime += get_runtime() - startTime;
  }
  
  
  
  void trainDataset::ReadBinaryData(std::string& data_filename){
	  long startTime = get_runtime();
	  std::ifstream data_file(data_filename.c_str(), std::ios::in | std::ios::binary);
	  if (!data_file.good()) {
		    std::cerr << "error w/ " << data_filename << std::endl;
		    exit(EXIT_FAILURE);
	  } 
	   data_file.seekg (0, data_file.end);
	   long total_size = data_file.tellg();
	   data_file.seekg (0, data_file.beg);

	   int unit_size =sizeof(float);	   
	   for(int j=0; j<total_size/(dimension*unit_size); j++){
	  	 std::ostringstream buf; 
	  	
	  	 std::vector<float> vec(dimension,0.0);
	  	 data_file.read((char*)vec.data(), unit_size*dimension);
	  	 if (data_file.good()){
	  	      if(format=="sparse")
	  	      	std::copy(vec.begin(), vec.end(), std::ostream_iterator<float>(buf, " "));
	  	      else{
	  	    	dense_vector instance; 
	  	        instance.initialize(dimension, vec);
	  	        double isn=instance.snorm();
	  	        if(normalize){
	  	         //instance.scale(1.0/sqrt(isn));
	  	         instance.normalize();
	  	         R=1.0;
	  	         Q.push_back(1.0);
	  	         }else{
	  	         if(isn>R) R=isn;
	  	         Q.push_back(isn);
	  	        }
	  	     	densedata.push_back(instance);
	  	      }
	  	 }
	  	 else { std::cout << "read data "<<j+1<<" failed"<<std::endl; exit(EXIT_FAILURE);}            

	     //std::cout<<buf.str()<<std::endl;
	  	 if(format=="sparse"){
	  	   std::istringstream is(buf.str());
	  	   simple_sparse_vector instance(is, dimension, false);
    	   double isn=instance.snorm();
 	       if(normalize){
 	        instance.scale(1.0/sqrt(isn));
 	        R=1.0;
 	        Q.push_back(1.0);
 	        }else{
 	         if(isn>R) R=isn;
 	         Q.push_back(isn);
 	       }
	  	   sparsedata.push_back(instance);
	  	 }
	     num_examples ++;
	   }
	   
	   data_file.close();
	   
       #ifdef nodef
          std::cerr << "num_examples = " << num_examples
	                << " dimension = " << dimension << std::endl;
       #endif
	   readingTime += get_runtime() - startTime;

  }
  
  
  void trainDataset::ReadLabels(std::string& label_filename){
	     long startTime = get_runtime();
		 std::ifstream label_file(label_filename.c_str());
		  if (!label_file.good()) {
		    std::cerr << "error w/ " << label_filename << std::endl;
		    exit(EXIT_FAILURE);
		  }
	    std::string line;
	    ulong n=0;
	    while(std::getline(label_file, line))
	    {
	      std::istringstream in(line);
	      if(task=="class"){
	        int label; 
	        in>>label;
	        S +=label*label;
	        uLabels.insert(label);
	        Labels.push_back(label);
	      }else{
	    	double label;
	    	in >> label;
	    	S +=label*label;
	    	Targets.push_back(label);
	      }
	    }
	    if(task=="class") n=Labels.size();
	    else  n = Targets.size();
	    if(n!=num_examples){
		    std::cerr << "error data size "<<num_examples<<" != label size "<<n<<" in "<<label_filename<< std::endl;
		    exit(EXIT_FAILURE);
	    }
	    	
	    label_file.close();
	   	// update timeline
	   	readingTime += get_runtime() - startTime;
  }
