// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013  Tianbao Yang.
//  All Rights Reserved.
//=============================================================================
// File Name: inc_dual.cc
// implements different updates in sdca for different loss functions
//=============================================================================
#include "inc_dual.h"


void dual_update(int ls, trainDataset& Data, WeightVector& W, WeightVector& WA, std::vector<double>& Alpha, 
		         double coeff,  std::vector<double>& phi_loss_error_norm, uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass)
{
   if(ls==1)      L1SVM_dual_update(Data, W, WA, Alpha, coeff, phi_loss_error_norm,  max_iter, exam_per_iter, index,  shift, myclass);
   else if(ls==2) L2SVM_dual_update(Data, W, WA, Alpha, coeff, phi_loss_error_norm,  max_iter, exam_per_iter, index,  shift, myclass);
   else if(ls==3) LogReg_dual_update(Data, W, WA, Alpha, coeff, phi_loss_error_norm,  max_iter, exam_per_iter, index,  shift,myclass);
   else if(ls==4) LestReg_dual_update(Data, W, WA, Alpha, coeff, phi_loss_error_norm,  max_iter, exam_per_iter, index,  shift);
}

void dual_update(int ls, trainDataset& Data, std::vector<int>& my_Class,  std::vector<WeightVector>& Wc, std::vector<WeightVector>& WAc, 
		         std::vector<std::vector<double> >& Alphac, double coeff,  std::vector<double>& phi_loss_norm, uint max_iter, uint exam_per_iter, 
		         std::vector<ulong>& index, ulong& shift)
{
	 if(ls==1)      L1SVM_dual_update(Data, my_Class, Wc, WAc, Alphac, coeff, phi_loss_norm,  max_iter, exam_per_iter, index,  shift);
	 else if(ls==2) L2SVM_dual_update(Data, my_Class, Wc, WAc, Alphac, coeff, phi_loss_norm,  max_iter, exam_per_iter, index,  shift);
	 else if(ls==3) LogReg_dual_update(Data, my_Class, Wc, WAc, Alphac, coeff, phi_loss_norm,  max_iter, exam_per_iter, index,  shift);

}



void L1SVM_dual_update(trainDataset& Data, WeightVector& W, WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		               uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass)
{
	      ulong ns=Data.size();
	      WeightVector RegW(W);
	      RegW.minus(WA, 1.0);
	      WA=W;
		  for(uint iter=0; iter<max_iter; iter++){

			  if(exam_per_iter==ns) std::random_shuffle(index.begin(), index.end()); //option I

			   for(uint r=0; r<exam_per_iter; r++){
				    ulong j;
				    if(exam_per_iter==ns)
				       j = index[r];	//option I
				    else
				       j = ((ulong)rand()) % ns; //options III
				    /*
			    	if(shift>=ns) {
			    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
			    	 	 shift=0;
			    	 }
			    	 j=index[shift]; 
			    	 shift++;*/ //options II
				    int is_class = (Data.y<int>(j)==myclass)?1:-1;

					double prediction = is_class*(Data.Xw(WA,j));
				    double v = (1 - prediction)/(Data.mQ(j)*coeff) + Alpha[j];
				  	v=std::max(std::min(v,1.0), 0.0);
				  	double del_alpha=v- Alpha[j];
				  	if(Data.fmt()=="dense") WA.add(Data.dX(j), del_alpha*is_class*coeff);
				  	else                    WA.add(Data.sX(j), del_alpha*is_class*coeff);
				  	phi[0] += del_alpha;
				  	Alpha[j]=v;
			   }
		   }
	       WA.minus(RegW,1.0);
}

void L2SVM_dual_update(trainDataset& Data, WeightVector& W, WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		               uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass)
{ 
       ulong ns=Data.size();
       WeightVector RegW(W);
       RegW.minus(WA, 1.0);
       WA=W;
	   for(uint iter=0; iter<max_iter; iter++){
			  if(exam_per_iter==ns)
			    std::random_shuffle(index.begin(), index.end()); //option I

			   for(uint r=0; r<exam_per_iter; r++){
				   ulong j;
				   if(exam_per_iter==ns)
				      j = index[r];	//option I
				   else
				      j = ((uint)rand()) % ns; //options III

				    /*
			    	if(shift>=ns) {
			    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
			    	 	 shift=0;
			    	 }
			    	 j=index[shift];
			    	 shift++; */ //options II
				     int is_class = (Data.y<int>(j)==myclass)?1:-1;
				   
					 double  prediction = is_class*(Data.Xw(WA, j));
					 double v = (1 - prediction - 0.5*Alpha[j])/(Data.mQ(j)*coeff +0.5) + Alpha[j];
				     v=std::max(v, 0.0);
				  	 double del_alpha=v- Alpha[j];
				  	 double sum_alpha=v+ Alpha[j];
				  	 if(Data.fmt()=="dense") WA.add(Data.dX(j), del_alpha*is_class*coeff);
				  	 else                    WA.add(Data.sX(j), del_alpha*is_class*coeff);
				  	 phi[0] += del_alpha*(1 - sum_alpha/4);
				  	 Alpha[j] =v;
			   }
		   }
	       WA.minus(RegW,1.0);	
}

void LogReg_dual_update(trainDataset& Data, WeightVector& W, WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		                uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass)
{ 
       ulong ns=Data.size();
       WeightVector RegW(W);
       RegW.minus(WA, 1.0);
       WA=W;
 	   for(uint iter=0; iter<max_iter; iter++){
 		  if(exam_per_iter==ns)
 		    std::random_shuffle(index.begin(), index.end()); //option I

 		   for(uint r=0; r<exam_per_iter; r++){
 			   ulong j;
 			   if(exam_per_iter==ns)
 			      j = index[r];	//option I
 			   else
 			      j = ((uint)rand()) % ns; //options III

 			    /*
 		    	if(shift>=ns) {
 		    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
 		    	 	 shift=0;
 		    	 }
 		    	 j=index[shift];
 		    	 shift++; */ //options II
			     int is_class = (Data.y<int>(j)==myclass)?1:-1;
 			   
 				 double  prediction = is_class*(Data.Xw(WA,j));
 			     double coeff2=coeff*coeff;
 			     double ix2 = 1.0/(0.25*coeff*Data.mQ(j) + 1);

 			     double del_alpha= (1.0/(1.0+exp(prediction))-Alpha[j])*ix2;
 			     del_alpha += ( 1.0/(1.0+exp(prediction+del_alpha*ix2*coeff2))-(Alpha[j]+del_alpha))*ix2;
 			     if (Alpha[j]+del_alpha<=1e-10) del_alpha=1e-10-Alpha[j];
 			     if (Alpha[j]+del_alpha>=1-(1e-10)) del_alpha= 1.0 - Alpha[j];

 			  	 double v= Alpha[j] + del_alpha;
 			  	 if(Data.fmt()=="dense") WA.add(Data.dX(j), del_alpha*is_class*coeff);
 			  	 else                    WA.add(Data.sX(j), del_alpha*is_class*coeff);
 			  	 phi[0] += (Alpha[j]*log(Alpha[j]+1e-12)+(1.0-Alpha[j])*log(1.0-Alpha[j]+(1e-12)));
 			  	 phi[0] -= (v*log(v+1e-12)+(1.0-v)*log(1.0-v+(1e-12)));
 			  	 Alpha[j] =v;
 		   }
 	   }
        WA.minus(RegW,1.0);
}

void LestReg_dual_update(trainDataset& Data, WeightVector& W, WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		                 uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift)
{
    ulong ns=Data.size();
    WeightVector RegW(W);
    RegW.minus(WA, 1.0);
    WA=W;
    for(uint iter=0; iter<max_iter; iter++){
   		  if(exam_per_iter==ns)
   		    std::random_shuffle(index.begin(), index.end()); //option I

   		   for(uint r=0; r<exam_per_iter; r++){
   			   ulong j;
   			   if(exam_per_iter==ns)
   			      j = index[r];	//option I
   			   else
   			      j = ((uint)rand()) % ns; //options III

   			    /*
   		    	if(shift>=ns) {
   		    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
   		    	 	 shift=0;
   		    	 }
   		    	 j=index[shift];
   		    	 shift++; */ //options II
   				 double  prediction = Data.Xw(WA, j);
   			     double  del_alpha = (Data.y<double>(j)-Alpha[j]- prediction)/(coeff*Data.mQ(j) + 1.0);
   			  	 double  sum_alpha = 2*Alpha[j] + del_alpha;
   			  	 if(Data.fmt()=="dense") WA.add(Data.dX(j), del_alpha*coeff);
   			  	 else                    WA.add(Data.sX(j), del_alpha*coeff);
   			  	 phi[0] +=del_alpha*(Data.y<double>(j) - sum_alpha/2.0);
   			  	 Alpha[j] +=del_alpha;
   		   }
   	   }
     WA.minus(RegW,1.0);
}


void L1SVM_dual_update(trainDataset& Data, std::vector<int>& my_Class, std::vector<WeightVector>& Wc, std::vector<WeightVector>& WAc,  
		               std::vector<std::vector<double> >& Alphac, double coeff, std::vector<double>& phi, uint max_iter, uint exam_per_iter, 
		               std::vector<ulong>& index, ulong& shift)
{
	    uint nc = my_Class.size();
	    ulong ns =Data.size();
     	std::vector<WeightVector> RegWc;
		RegWc.reserve(nc);
		for(uint c=0; c<nc; c++){
		 WeightVector RegW(Wc[c]);
		 RegW.minus(WAc[c], 1.0);
		 RegWc.push_back(RegW);
		 WAc[c]=Wc[c];
		}
		  for(uint iter=0; iter<max_iter; iter++){

			  if(exam_per_iter==ns)
			     std::random_shuffle(index.begin(), index.end()); //option I

			   for(uint r=0; r<exam_per_iter; r++){
				    ulong j;
				    if(exam_per_iter==ns)
				       j = index[r];	//option I
				    else
				       j = ((ulong)rand()) % ns; //options III
				    /*
			    	if(shift>=ns) {
			    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
			    	 	 shift=0;
			    	 }
			    	 j=index[shift]; 
			    	 shift++*/ //options II
	                for (uint c=0; c<nc; c++){
				      int is_class=(Data.y<int>(j)==my_Class[c])?1:-1;
					  double prediction = is_class*Data.Xw(WAc[c],j);
					  double alpha =Alphac[c][j];
				      //double v = (1 - prediction)/(Q[j]*coeff) + Alphac[c][j];
				      double v = (1 - prediction)/(Data.mQ(j)*coeff) + alpha;
				  	  v=std::max(std::min(v,1.0), 0.0);
				  	  //double del_alpha=v- Alphac[c][j];
				  	  double del_alpha=v - alpha;
				  	  if(Data.fmt()=="dense") WAc[c].add(Data.dX(j), del_alpha*is_class*coeff);
				  	  else                    WAc[c].add(Data.sX(j), del_alpha*is_class*coeff);
				  	  phi[c] += del_alpha;
				  	  Alphac[c][j]=v;
	                }
			   }
		   }
	      for(uint c=0; c<nc; c++){
	        WAc[c].minus(RegWc[c], 1.0);
	      }
}


void L2SVM_dual_update(trainDataset& Data, std::vector<int>& my_Class, std::vector<WeightVector>& Wc, std::vector<WeightVector>& WAc,  
		               std::vector<std::vector<double> >& Alphac, double coeff, std::vector<double>& phi, uint max_iter, uint exam_per_iter, 
		               std::vector<ulong>& index, ulong& shift)
{
	    uint nc = my_Class.size();
	    ulong ns =Data.size();
     	std::vector<WeightVector> RegWc;
		RegWc.reserve(nc);
		for(uint c=0; c<nc; c++){
		 WeightVector RegW(Wc[c]);
		 RegW.minus(WAc[c], 1.0);
		 RegWc.push_back(RegW);
		 WAc[c]=Wc[c];
		}
		  for(uint iter=0; iter<max_iter; iter++){
			  if(exam_per_iter==ns)
			    std::random_shuffle(index.begin(), index.end()); //option I

			   for(uint r=0; r<exam_per_iter; r++){
				   ulong j;
				   if(exam_per_iter==ns)
				      j = index[r];	//option I
				   else
				      j = ((ulong)rand()) % ns; //options III

				    /*
			    	if(shift>=ns) {
			    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
			    	 	 shift=0;
			    	 }
			    	 j=index[shift];
			    	 (shift)++; */ //options II
				   for(uint c=0; c<nc; c++){
					 int is_class=(Data.y<int>(j)==my_Class[c])?1:-1;
					 double  prediction = is_class*Data.Xw(WAc[c], j);
					 double alpha=Alphac[c][j];
					 //double v = (1 - prediction - 0.5*Alphac[c][j])/(Q[j]*coeff +0.5) + Alphac[c][j];
					 double v = (1 - prediction - 0.5*Alphac[c][j])/(Data.mQ(j)*coeff +0.5) + alpha;
				     v=std::max(v, 0.0);
				  	 //double del_alpha=v- Alphac[c][j];
				  	 //double sum_alpha=v+ Alphac[c][j];
				  	 double del_alpha=v- alpha;
				  	 double sum_alpha=v+ alpha;
				  	 if(Data.fmt()=="dense")  WAc[c].add(Data.dX(j), del_alpha*is_class*coeff);
				  	 else WAc[c].add(Data.sX(j), del_alpha*is_class*coeff);

				  	 phi[c] += del_alpha*(1 - sum_alpha/4);
				  	 Alphac[c][j] =v;
				   }
			   }
		   }
	      for(uint c=0; c<nc; c++){
	        WAc[c].minus(RegWc[c], 1.0);
	      }
}


void LogReg_dual_update(trainDataset& Data, std::vector<int>& my_Class, std::vector<WeightVector>& Wc, std::vector<WeightVector>& WAc,  
		                std::vector<std::vector<double> >& Alphac, double coeff, std::vector<double>& phi, uint max_iter, uint exam_per_iter, 
		                std::vector<ulong>& index, ulong& shift)
{
	    uint nc = my_Class.size();
	    ulong ns =Data.size();
     	std::vector<WeightVector> RegWc;
		RegWc.reserve(nc);
		for(uint c=0; c<nc; c++){
		 WeightVector RegW(Wc[c]);
		 RegW.minus(WAc[c], 1.0);
		 RegWc.push_back(RegW);
		 WAc[c]=Wc[c];
		}
		  for(uint iter=0; iter<max_iter; iter++){
			  if(exam_per_iter==ns)
			    std::random_shuffle(index.begin(), index.end()); //option I

			   for(uint r=0; r<exam_per_iter; r++){
				   ulong j;
				   if(exam_per_iter==ns)
				      j = index[r];	//option I
				   else
				      j = ((ulong)rand()) % ns; //options III

				    /*
			    	if(shift>=ns) {
			    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
			    	 	 shift=0;
			    	 }
			    	 j=index[shift];
			    	 (shift)++; */ //options II

				   for(uint c=0; c<nc; c++){
					 int is_class=(Data.y<int>(j)==my_Class[c])?1:-1;
					 double  prediction = is_class*Data.Xw(WAc[c], j);
				     double coeff2=coeff*coeff;
				     double ix2 = 1.0/(0.25*coeff*Data.mQ(j) + 1);

				     double alpha=Alphac[c][j];
				     
				     /*double del_alpha= (1.0/(1.0+exp(prediction))-Alphac[c][j])*ix2;
				     del_alpha += ( 1.0/(1.0+exp(prediction+del_alpha*ix2*coeff2))-(Alpha[c][j]+del_alpha))*ix2;
				     if (Alphac[c][j]+del_alpha<=1e-10) del_alpha=1e-10-Alphac[c][j];
				     if (Alphac[c][j]+del_alpha>=1-(1e-10)) del_alpha= 1.0 - Alphac[c][j];
				  	 double v= Alphac[c][j] + del_alpha;
				  	 WAc[c].add(Dataset[j], del_alpha*is_class*coeff);
				  	 *phi += (Alphac[c][j]*log(Alphac[c][j]+1e-12)+(1.0-Alphac[c][j])*log(1.0-Alphac[c][j]+(1e-12)));
				  	 *phi -= (v*log(v+1e-12)+(1.0-v)*log(1.0-v+(1e-12)));*/
				     
				     double del_alpha= (1.0/(1.0+exp(prediction))-alpha)*ix2;
					 del_alpha += ( 1.0/(1.0+exp(prediction+del_alpha*ix2*coeff2))-(alpha+del_alpha))*ix2;
					 if (alpha+del_alpha<=1e-10) del_alpha=1e-10-alpha;
					 if (alpha+del_alpha>=1-(1e-10)) del_alpha= 1.0 - alpha;
					 double v= alpha + del_alpha;
					 if(Data.fmt()=="dense") WAc[c].add(Data.dX(j), del_alpha*is_class*coeff);
					 else                    WAc[c].add(Data.sX(j), del_alpha*is_class*coeff);
					 phi[c] += (alpha*log(alpha+1e-12)+(1.0-alpha)*log(1.0-alpha+(1e-12)));
					 phi[c] -= (v*log(v+1e-12)+(1.0-v)*log(1.0-v+(1e-12)));		  	 
				  	 
				  	 Alphac[c][j] =v;
				   }
			   }
		   }
	      for(uint c=0; c<nc; c++){
	        WAc[c].minus(RegWc[c], 1.0);
	      }
}

//----------------------------- for communication efficient optimization --------------------------------------------------
void dual_epoch_update(int ls, trainDataset& Data,  WeightVector& WA, std::vector<double>& Alpha, 
		         double coeff,  std::vector<double>& phi_loss_error_norm, uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass)
{
   if(ls==1)      L1SVM_dual_epoch_update(Data,   WA, Alpha,  coeff, phi_loss_error_norm,  max_iter, exam_per_iter, index,  shift, myclass);
   else if(ls==2) L2SVM_dual_epoch_update(Data,   WA, Alpha,  coeff, phi_loss_error_norm,  max_iter, exam_per_iter, index,  shift, myclass);
   else if(ls==3) LogReg_dual_epoch_update(Data,  WA, Alpha,  coeff, phi_loss_error_norm,  max_iter, exam_per_iter, index,  shift,myclass);
   else if(ls==4) LestReg_dual_epoch_update(Data, WA, Alpha,  coeff, phi_loss_error_norm,  max_iter, exam_per_iter, index,  shift);
}


void L1SVM_dual_epoch_update(trainDataset& Data,  WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		               uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass)
{
	      ulong ns=Data.size();
		  for(uint iter=0; iter<max_iter; iter++){

			  if(exam_per_iter==ns) std::random_shuffle(index.begin(), index.end()); //option I

			   for(uint r=0; r<exam_per_iter; r++){
				    ulong j;
				    if(exam_per_iter==ns)
				       j = index[r];	//option I
				    else
				       j = ((ulong)rand()) % ns; //options III
				    /*
			    	if(shift>=ns) {
			    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
			    	 	 shift=0;
			    	 }
			    	 j=index[shift]; 
			    	 shift++;*/ //options II
				    int is_class = (Data.y<int>(j)==myclass)?1:-1;

					double prediction = is_class*(Data.Xw(WA,j));
				    double v = (1 - prediction )/(Data.mQ(j)*coeff) + Alpha[j];
				  	v=std::max(std::min(v,1.0), 0.0);
				  	double del_alpha=v- Alpha[j];
				  	if(Data.fmt()=="dense") WA.add(Data.dX(j), del_alpha*is_class*coeff);
				  	else                    WA.add(Data.sX(j), del_alpha*is_class*coeff);
				  	phi[0] += del_alpha;
				  	Alpha[j]=v;
			   }
		   }
}

void L2SVM_dual_epoch_update(trainDataset& Data, WeightVector& WA, std::vector<double>& Alpha,  double coeff,  std::vector<double>& phi, 
		               uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass)
{ 
       ulong ns=Data.size();
	   for(uint iter=0; iter<max_iter; iter++){
			  if(exam_per_iter==ns)
			    std::random_shuffle(index.begin(), index.end()); //option I

			   for(uint r=0; r<exam_per_iter; r++){
				   ulong j;
				   if(exam_per_iter==ns)
				      j = index[r];	//option I
				   else
				      j = ((uint)rand()) % ns; //options III

				    /*
			    	if(shift>=ns) {
			    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
			    	 	 shift=0;
			    	 }
			    	 j=index[shift];
			    	 shift++; */ //options II
				     int is_class = (Data.y<int>(j)==myclass)?1:-1;
				   
					 double  prediction = is_class*(Data.Xw(WA, j));
					 double v = (1 - prediction - 0.5*Alpha[j])/(Data.mQ(j)*coeff +0.5) + Alpha[j];
				     v=std::max(v, 0.0);
				  	 double del_alpha=v- Alpha[j];
				  	 double sum_alpha=v+ Alpha[j];
				  	 if(Data.fmt()=="dense") WA.add(Data.dX(j), del_alpha*is_class*coeff);
				  	 else                    WA.add(Data.sX(j), del_alpha*is_class*coeff);
				  	 phi[0] += del_alpha*(1 - sum_alpha/4);
				  	 Alpha[j] =v;
			   }
		   }
}

void LogReg_dual_epoch_update(trainDataset& Data,  WeightVector& WA, std::vector<double>& Alpha,  double coeff,  std::vector<double>& phi, 
		                uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass)
{ 
       ulong ns=Data.size();
 	   for(uint iter=0; iter<max_iter; iter++){
 		  if(exam_per_iter==ns)
 		    std::random_shuffle(index.begin(), index.end()); //option I

 		   for(uint r=0; r<exam_per_iter; r++){
 			   ulong j;
 			   if(exam_per_iter==ns)
 			      j = index[r];	//option I
 			   else
 			      j = ((uint)rand()) % ns; //options III

 			    /*
 		    	if(shift>=ns) {
 		    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
 		    	 	 shift=0;
 		    	 }
 		    	 j=index[shift];
 		    	 shift++; */ //options II
			     int is_class = (Data.y<int>(j)==myclass)?1:-1;
 			   
 				 double  prediction = is_class*(Data.Xw(WA,j));
 			     double coeff2=coeff*coeff;
 			     double ix2 = 1.0/(0.25*coeff*Data.mQ(j) + 1);

 			     double del_alpha= (1.0/(1.0+exp(prediction))-Alpha[j])*ix2;
 			     del_alpha += ( 1.0/(1.0+exp(prediction+del_alpha*ix2*coeff2))-(Alpha[j]+del_alpha))*ix2;
 			     if (Alpha[j]+del_alpha<=1e-10) del_alpha=1e-10-Alpha[j];
 			     if (Alpha[j]+del_alpha>=1-(1e-10)) del_alpha= 1.0 - Alpha[j];

 			  	 double v= Alpha[j] + del_alpha;
 			  	 if(Data.fmt()=="dense") WA.add(Data.dX(j), del_alpha*is_class*coeff);
 			  	 else                    WA.add(Data.sX(j), del_alpha*is_class*coeff);
 			  	 phi[0] += (Alpha[j]*log(Alpha[j]+1e-12)+(1.0-Alpha[j])*log(1.0-Alpha[j]+(1e-12)));
 			  	 phi[0] -= (v*log(v+1e-12)+(1.0-v)*log(1.0-v+(1e-12)));
 			  	 Alpha[j] =v;
 		   }
 	   }
}

void LestReg_dual_epoch_update(trainDataset& Data, WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		                 uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift)
{
    ulong ns=Data.size();
    for(uint iter=0; iter<max_iter; iter++){
   		  if(exam_per_iter==ns)
   		    std::random_shuffle(index.begin(), index.end()); //option I

   		   for(uint r=0; r<exam_per_iter; r++){
   			   ulong j;
   			   if(exam_per_iter==ns)
   			      j = index[r];	//option I
   			   else
   			      j = ((uint)rand()) % ns; //options III

   			    /*
   		    	if(shift>=ns) {
   		    	 	 std::random_shuffle(index.begin(), index.end()); // implemented by random permutation  (is better than random sampling)
   		    	 	 shift=0;
   		    	 }
   		    	 j=index[shift];
   		    	 shift++; */ //options II
   				 double  prediction = Data.Xw(WA, j);
   			     double  del_alpha = (Data.y<double>(j)-Alpha[j]- prediction)/(coeff*Data.mQ(j) + 1.0);
   			  	 double  sum_alpha = 2*Alpha[j] + del_alpha;
   			  	 if(Data.fmt()=="dense") WA.add(Data.dX(j), del_alpha*coeff);
   			  	 else                    WA.add(Data.sX(j), del_alpha*coeff);
   			  	 phi[0] +=del_alpha*(Data.y<double>(j) - sum_alpha/2.0);
   			  	 Alpha[j] +=del_alpha;
   		   }
   	   }
}
