// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013  Tianbao Yang.
//  All Rights Reserved.
//=============================================================================
// File Name: inc_dual.h
// head file for different updates in sdca for different loss functions
//=============================================================================

#ifndef INC_DUAL_H_
#define INC_DUAL_H_

//*****************************************************************************
// Included Files
//*****************************************************************************
#include <iostream>
#include <ostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <iterator>
#include <vector>
#include <sstream>
#include <ctime>
#include <cmath>
#include <algorithm>
#include "simple_sparse_vec_hash.h"
#include "WeightVector.h"
#include "data_set.h"
#define INF HUGE_VAL

void dual_update(int ls, trainDataset& Data, WeightVector& W, WeightVector& Wr, std::vector<double>& Alpha, 
		         double coeff,  std::vector<double>& phi_loss_error_norm, uint max_iter, uint exam_per_iter,
		         std::vector<ulong>& index, ulong& shift, int myclass=1);	

void dual_update(int ls, trainDataset& Data, std::vector<int>& my_Class,  std::vector<WeightVector>& Wc, std::vector<WeightVector>& Wrc, 
		         std::vector<std::vector<double> >& Alphac, double coeff,  std::vector<double>& phi_loss_norm, uint max_iter, uint exam_per_iter, 
		         std::vector<ulong>& index, ulong& shift);


void L1SVM_dual_update(trainDataset& Data, WeightVector& W, WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		               uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass=1);

void L2SVM_dual_update(trainDataset& Data, WeightVector& W, WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		               uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass=1);

void LogReg_dual_update(trainDataset& Data, WeightVector& W, WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		               uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass=1);

void LestReg_dual_update(trainDataset& Data, WeightVector& W, WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		               uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift);


void L1SVM_dual_update(trainDataset& Data, std::vector<int>& my_Class, std::vector<WeightVector>& RegWc, std::vector<WeightVector>& WAc,  
		               std::vector<std::vector<double> >& Alphac, double coeff, std::vector<double>& phi, uint max_iter, uint exam_per_iter, 
		               std::vector<ulong>& index, ulong& shift);

void L2SVM_dual_update(trainDataset& Data, std::vector<int>& my_Class, std::vector<WeightVector>& RegWc, std::vector<WeightVector>& WAc,  
		               std::vector<std::vector<double> >& Alphac, double coeff, std::vector<double>& phi, uint max_iter, uint exam_per_iter, 
		               std::vector<ulong>& index, ulong& shift);

void LogReg_dual_update(trainDataset& Data, std::vector<int>& my_Class, std::vector<WeightVector>& RegWc, std::vector<WeightVector>& WAc,  
		               std::vector<std::vector<double> >& Alphac, double coeff, std::vector<double>& phi, uint max_iter, uint exam_per_iter, 
		               std::vector<ulong>& index, ulong& shift);


//----------------------------- for communication efficient optimization --------------------------------------------------
void dual_epoch_update(int ls, trainDataset& Data,  WeightVector& WA, std::vector<double>& Alpha,  
		         double coeff,  std::vector<double>& phi_loss_error_norm, uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass);

void L1SVM_dual_epoch_update(trainDataset& Data,  WeightVector& WA, std::vector<double>& Alpha,  double coeff,  std::vector<double>& phi, 
		               uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass);

void L2SVM_dual_epoch_update(trainDataset& Data, WeightVector& WA, std::vector<double>& Alpha,   double coeff,  std::vector<double>& phi, 
		               uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass);

void LogReg_dual_epoch_update(trainDataset& Data,  WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		                uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift, int myclass);

void LestReg_dual_epoch_update(trainDataset& Data, WeightVector& WA, std::vector<double>& Alpha, double coeff,  std::vector<double>& phi, 
		                 uint max_iter, uint exam_per_iter, std::vector<ulong>& index, ulong& shift);


#endif /*INC_DUAL_H_*/
